""" Classes to hold grid information


Classes
-------
UniformGrid : grid info.
              Stores information on the 2d grid in spherical polar coordinates.
GLGrid : grid info
         Stores a Gauss-Legendre type grid on a spherical surface.
"""

import numpy as np

# from numba import jit, njit

# import numba as nb

# from numba.experimental import jitclass


class UniformGrid:
    """A class to store the theta-phi grid info."""

    def __init__(
        self,
        nphi=80,
        ntheta=41,
        nphimax=124,
        nthetamax=66,
        nghosts=2,
        integration_method="MP",
    ):

        # Number of gridpoints along phi direction including ghost points.
        self.nphi = nphi
        # Number of gridpoints along theta direction including ghost points.
        self.ntheta = ntheta
        # Total length of phi array used by ETK.
        self.nphimax = nphimax
        # Total length of theta array used by ETK.
        self.nthetamax = nthetamax
        # Number of ghost points in theta/phi direction.
        self.nghosts = nghosts
        # The default integration method
        self._integration_method = integration_method

    @property
    def npix(self):
        # Return the total number of pixels, including the ghost zones present at one iteration.
        return (self.ntheta) * (self.nphi)

    @property
    def npix_act(self):
        # Return the actual number of pixels, excluding the ghost zones present at one iteration.
        return (self.ntheta - 2 * self.nghosts) * (
            self.nphi - 2 * self.nghosts
        )

    @property
    def npix_max(self):
        # Return the (max) total number of pixels, including the ghost and buffer zones at one iteration.
        return (self.nthetamax) * (self.nphimax)

    @property
    def ntheta_act(self):
        """Return the actual number of valid pixels,
        excluding the ghost and buffer zones, along the
        theta axis at one iteration."""
        return self.ntheta - 2 * self.nghosts

    @property
    def nphi_act(self):
        """Return the actual number of valid pixels,
        excluding the ghost and buffer zones,
        along the phi axis at one iteration."""
        return self.nphi - 2 * self.nghosts

    @property
    def dtheta(self):
        # Return the coodinate spacing d\theta
        return np.pi / (self.ntheta - 2 * self.nghosts)

    @property
    def dphi(self):
        # Return the coordinate spacing d\phi
        return 2 * np.pi / (self.nphi - 2 * self.nghosts)

    @property
    def nbuffer(self):
        # Return the coordinate spacing d\phi
        return self.nthetamax - self.ntheta

    @property
    def shape(self):
        """Return the shape of the grif"""
        return (self.ntheta_act, self.nphi_act)

    @property
    def theta_1d(self, theta_index=None):
        """Returns the coordinate value theta 
        given the coordinate index. The coordinate 
        index ranges from (0, ntheta). The actual 
        indices without the ghost and extra zones 
        is (nghosts, ntheta-nghosts).

        Parameters
        -----------
        theta_index : int/ 1d array
                      The theta coordinate index or axis.

        Returns
        -------
        theta_1d : float
                   The coordinate(s) :math:`\\theta` on the sphere.
        """

        if not theta_index:
            theta_index = np.arange(self.nghosts, self.ntheta - self.nghosts)

        return (
            (theta_index - self.nghosts + 0.5)
            * np.pi
            / (self.ntheta - 2 * self.nghosts)
        )

    @property
    def phi_1d(self, phi_index=None):
        """Returns the coordinate value theta given 
        the coordinate index. The coordinate index lies 
        in (0, nphi). The actual indices without 
        the ghost and extra zones is (nghosts, nphi-nghosts).

        Parameters
        -----------
        phi_1d : int / 1d array
                 The phi coordinate index or axis.

        Returns
        -------
        phi_1d : float or 1d array
                 The coordinate(s) :math:`\\phi` on the sphere.

        """

        if not phi_index:
            phi_index = np.arange(self.nghosts, self.nphi - self.nghosts)

        return (
            (phi_index - self.nghosts)
            * 2
            * np.pi
            / (self.nphi - 2 * self.nghosts)
        )

    @property
    def meshgrid(self):
        """The (:math:`\\theta, \\phi)`: coordinate meshes.
        Excludes the ghost zones.

        Returns
        -------
        theta :	2d array
                The :math:`\\theta` coordinate matrix 
                for vectorization.

        phi : 2d array
              The :math:`\\phi` coordinate matrix 
              for vectorization.
        """

        theta, phi = np.meshgrid(self.theta_1d, self.phi_1d)

        # theta = np.zeros((self.ntheta_act, self.nphi_act))
        # phi	= np.zeros((self.ntheta_act, self.nphi_act))

        # for theta_index, theta_val in enumerate(self.theta_1d()):
        # 	for phi_index, phi_val in enumerate(self.phi_1d()):
        # 		theta[theta_index, phi_index] = theta_val
        # 		phi[theta_index, phi_index] = phi_val

        return np.transpose(theta), np.transpose(phi)

    @property
    def integration_method(self):
        """The default integration method"""
        return self._integration_method


class GLGrid:
    """A class to store the coordinate grid on a sphere.

    Attributes
    ----------
    ntheta : int
             The number of angular points in the :math:`\\theta`
             direction, including ghost zones.
    nphi : int
           The number of angular points in the :math:`\\phi`
           direction, including ghost zones.
    nghosts : int
              The number of ghost zones at the end of
              each direction.
    meshgrid : tuple of 2d array
               The 2d array containing the meshgrid of
               (:math:`\\theta, \\phi`) angular points.
    theta_1d : 1d array
               The 1d array of angular points
               along the :math:`\\theta` axis.
    phi_1d : 1d array
             The 1d array of angular points
             along the :math:`\\phi` axis.
    dtheta : float
             The angular step size in the :math:`\\theta`
             direction.
    dphi : float
           The angular step size inthe :math:`\\phi`
           direction.
    npix_act : int
               The total number of gridpoints on the sphere,
               excluding the ghost points.
    meshgrid : tuple of 2darray
               Get the 2d angular grid.

    Methods
    -------
    theta_1d :
        Get the :math:`\\theta` axis.
    phi_1d :
        Get the :math:`\\phi` axis.

    Notes
    -----
    The total number of points on the sphere
    is assumed to be :math:`2 (L+1)^2`

    :math:`N_\theta = L+1`

    :math:`N_\phi = 2(L+1)`

    This integrates out spherical harmonics of degree L exactly,
    given a regular function on the sphere.
    """

    def __init__(
        self,
        nphi=None,
        ntheta=None,
        nphi_act=None,
        ntheta_act=None,
        L=47,
        nghosts=2,
        integration_method="GL"):

        # Number of gridpoints along phi direction including ghost points.
        self._nphi = nphi
        # Number of gridpoints along theta direction including ghost points.
        self._ntheta = ntheta
        self._nphi_act = nphi_act
        self._ntheta_act = ntheta_act
        # Total length of phi array used by ETK.
        # self._nphimax		= nphimax
        # Total length of theta array used by ETK.
        # self._nthetamax	= nthetamax
        # Number of ghost points in theta/phi direction.
        self._nghosts = nghosts
        self._L = L
        self._theta_1d = None
        self._phi_1d = None
        self._meshgrid = None
        self._integration_method = integration_method

        if self._ntheta is None:
            if self._L is None:
                raise ValueError("Please specify L or angular points!")

            else:
                self._ntheta_act = L + 1
                self._nphi_act = 2 * self._ntheta_act

                self._ntheta = self._ntheta_act + 2 * self._nghosts
                self._nphi = self._nphi_act + 2 * self._nghosts

        elif self._L is None:
            if self.nthetha is None:
                raise ValueError("Please specify L or angular points!")
            else:
                self.L = self.ntheta_act - 1
                # assert(nphi%2==0)
                self.nphi_act = 2 * self.L + 1

        from scipy.special import roots_legendre

        cpoints, self._weights, self._sum_of_weights = roots_legendre(
            L + 1, mu=True
        )

        # xpoints = (np.pi-np.arccos(cpoints))
        xpoints = np.arccos(cpoints[::-1])
        self._theta_1d = xpoints

        dphi = 2 * np.pi / self._nphi_act

        self._phi_1d = np.linspace(0, 2 * np.pi - dphi, self._nphi_act)

        theta_grid, phi_grid = np.meshgrid(self._theta_1d, self._phi_1d)
        self._meshgrid = np.transpose(theta_grid), np.transpose(phi_grid)

        dtheta_axis = np.diff(self._theta_1d)

        dtheta_axis = np.append(dtheta_axis, np.pi - self._theta_1d[-1])
        dtheta_axis = np.insert(dtheta_axis, 0, self._theta_1d[0])

        self._dtheta_1d = dtheta_axis

        self._dphi = dphi  # self._phi_1d[1]

    def nphi(self):
        """Return the total number of gridpoints
        along the phi direction, including
        the ghost zones at one iteration."""
        if not self._nphi:
            return self._nphi_act + self.nghosts
        else:
            return self._nphi

    def ntheta(self):
        """Return the total number gridpoints
        along the theta direction, including
        the ghost zones at one iteration."""
        if not self._ntheta:
            return self._ntheta_act + self.nghosts
        else:
            return self._ntheta

    @property
    def nghosts(self):
        """Return the number of ghost zones
        at each end in each direction"""
        return self._nghosts

    @property
    def ntheta_act(self):
        """Return the actual number of physical data points
        along the theta direction excluding the ghost and
        buffer zones at one iteration."""
        if not self._ntheta_act:
            return self.ntheta - 2 * self.nghosts
        else:
            return self._ntheta_act

    @property
    def nphi_act(self):
        """Return the actual number of physical data points
        along the phi direction excluding the ghost and
        buffer zones at one iteration."""

        if not self._ntheta_act:
            return self.nphi - 2 * self.nghosts
        else:
            return self._nphi_act

    @property
    def npix_act(self):
        """Return the actual number of pixels,
        excluding the ghost zones present
        at one iteration"""
        return (self.ntheta_act) * (self.nphi_act)

    @property
    def dtheta_1d(self):
        """Return the non-uniform angular stepping in 
        :math:`\theta` direction"""
        return self._dtheta_1d

    @property
    def dphi(self):
        """Return the uniform angular stepping in :math:`\phi` direction"""
        return self._dphi

    @property
    def L(self):
        """Return the total number of pixels, including 
        the ghost zones present at one iteration."""
        return self._L

    @property
    def npix(self):
        """Return the total number of pixels, including 
        the ghost zones present at one iteration."""
        return (self.ntheta) * (self.nphi)

    @property
    def weights(self):
        """Return the integration weights along theta"""
        return self._weights

    @property
    def weights_grid(self):
        """Return the integration weights on the coordinate mesh grid"""
        return np.outer(self.weights, np.ones(self.nphi_act))

    @property
    def shape(self):
        """Return the shape of the grif"""
        return self.weights_grid.shape

    @property
    def theta_1d(self):
        """Returns the coordinate value theta given the coordinate index. 
        The coordinate index ranges from (0, ntheta). The actual indices 
        without the ghost and extra zones is (nghosts, ntheta-nghosts).

        Parameters
        -----------
        theta_index : int/ 1d array
                      The theta coordinate index or axis.

        Returns
        -------
        theta_1d : float
                   The coordinate(s) :math:`\\theta` on the sphere.
        """

        return self._theta_1d

    @property
    def phi_1d(self):
        """Returns the coordinate value theta given the coordinate index. 
        The coordinate index lies in (0, nphi). The actual indices without 
        the ghost and extra zones is (nghosts, nphi-nghosts).

        Parameters
        -----------
        phi_1d : int / 1d array
                 The phi coordinate index or axis.

        Returns
        -------
        phi_1d : float or 1d array
                 The coordinate(s) :math:`\\phi` on the sphere.
        """

        return self._phi_1d

    @property
    def meshgrid(self):
        """The (:math:`\\theta, \\phi)`: coordinate meshes.
        Excludes the ghost zones.


        Returns
        -------
        theta :	2d array
                The :math:`\\theta` coordinate matrix for vectorization.

        phi : 2d array
              The :math:`\\phi` coordinate matrix for vectorization.
        """

        return self._meshgrid

    @property
    def integration_method(self):
        """The default integration method"""
        return self._integration_method



""" Methods to transform the waveform """

import numpy as np

from waveformtools.waveformtools import message

# from numba import jit, njit


# @njit(parallel=True)
def compute_fft(udata_x, delta_x):
    """Find the FFT of the samples in time-space, and return with the frequencies.

    Parameters
    ----------

    udata_x:	1d array
                                                    The samples in time-space.

    delta_x:	float
                                                    The stepping delta_x

    Returns
    -------

    freqs:	1d array
                                    The frequency axis, shifted approriately.
    utilde:	1d array
                                                    The samples in frequency space, with conventions applied.

    """

    # import necessary libraries.
    from numpy.fft import fft

    # FFT
    utilde_orig = fft(udata_x)

    # Apply conventions.
    utilde = set_fft_conven(utilde_orig)

    # Get frequency axes.
    Nlen = len(utilde)
    # message(Nlen)
    # Naxis			= np.arange(Nlen)
    # freq_orig		= fftfreq(Nlen)
    # freq_axis		= fftshift(freq_orig)*Nlen
    # delta_x		 = xdata[1] - xdata[0]

    # Naxis			 = np.arange(Nlen)
    #freq_axis = np.linspace(-0.5 / delta_x, 0.5 / delta_x, Nlen)
    freq_axis = np.fft.fftshift(np.fft.fftfreq(Nlen, delta_x))

    return freq_axis, utilde


# @njit(parallel=True)
def compute_ifft(utilde, delta_f):
    """Find the inverse FFT of the samples in frequency-space, and return with the time axis.

    Parameters
    ----------

    utilde	:	1d array
                                                    The samples in frequency-space.

    delta_f:	float
                                                    The frequency stepping

    Returns
    -------

    time_axis:	1d array
                                                    The time axis.

    udata_time:	1d array
                                                                    The samples in time domain.

    """

    # import necessary libraries.
    from numpy.fft import ifft

    # FFT
    utilde_orig = unset_fft_conven(utilde)

    # Inverse transform
    udata_time = ifft(utilde_orig)

    # Get frequency axes.
    Nlen = len(udata_time)
    # message(Nlen)
    # Naxis			= np.arange(Nlen)
    # freq_orig		= fftfreq(Nlen)
    # freq_axis		= fftshift(freq_orig)*Nlen
    # delta_x		 = xdata[1] - xdata[0]

    # Naxis			 = np.arange(Nlen)
    delta_t = 1.0 / (delta_f * Nlen)
    # Dt				= Nlen * delta_f/2

    # time_axis = np.linspace(0, delta_t * Nlen, Nlen)
    time_axis = np.arange(0, delta_t * Nlen, 1/Nlen)

    return time_axis, udata_time


# @njit(parallel=True)
def set_fft_conven(utilde_orig):
    """Make a numppy fft consistent with the chosen conventions.
                    This takes care of the zero mode factor and array position.
                    Also, it shifts the negative frequencies using numpy's fftshift.

    Parameters
    ----------

    utilde_orig:	1d array
                                                                    The result of a numpy fft.

    Returns
    -------

    utilde_conven:	1d array
                                                                    The fft with set conventions.
    """

    # Multiply by 2, take conjugate.
    utilde_conven = 2 * np.conj(utilde_orig) / len(utilde_orig)
    # Restore the zero mode.
    utilde_conven[0] = utilde_conven[0] / 2
    # Shift the frequency axis.
    utilde_conven = np.fft.fftshift(utilde_conven)

    return utilde_conven


# @njit(parallel=True)
def unset_fft_conven(utilde_conven):
    """Make an actual conventional fft consistent with numpy's conventions.
                    The inverse of set_conv.


    Parameters
    ----------

    utilde_conven:	1d array
                                                                    The conventional fft data vector.

    Returns
    -------

    utilde_np
    """

    utilde_np = np.fft.ifftshift(utilde_conven)

    utilde_np = len(utilde_np) * np.conj(utilde_np) / 2
    # message(utilde_original[0])
    utilde_np[0] *= 2
    # message(utilde_original[0])

    return utilde_np


def Yslm(spin_weight, ell, emm, theta, phi):
    """Spin-weighted spherical harmonics fast evaluation.

    Parameters
    ----------

    spin_weight :	int
                                    The Spin weight.
    ell :	int
                    The mode number :math:`\\ell'.
    emm :	int
                    The azimuthal mode number :math:`m'.
    theta : float
                    The polar angle  :math:`\\theta` in radians,
    phi :	float
                    The aximuthal angle :math:`\\phi' in radians.

    Returns
    --------
    Yslm :	float
                    The value of Yslm at :math:`\\theta, phi'.

            Note
            ----
            This is accurate upto 14 decimals for L upto 25.

    """
    import sympy as sp

    # theta, phi = sp.symbols('theta phi')

    fact = np.math.factorial
    # fact = sp.factorial
    Sum = 0

    factor = 1
    if spin_weight < 0:
        factor = (-1) ** ell
        theta = np.pi - theta
        phi += np.pi

    abs_spin_weight = abs(spin_weight)

    for aar in range(ell - abs_spin_weight + 1):
        if (aar + abs_spin_weight - emm) < 0 or (
            ell - aar - abs_spin_weight
        ) < 0:
            message(f"Skippin r {aar}", message_verbosity=3)
            continue
        else:
            Sum += (
                sp.binomial(ell - abs_spin_weight, aar)
                * sp.binomial(
                    ell + abs_spin_weight, aar + abs_spin_weight - emm
                )
                * np.power((-1), (ell - aar - abs_spin_weight))
                * np.exp(1j * emm * phi)
                / np.power(
                    np.tan(theta / 2), (2 * aar + abs_spin_weight - emm)
                )
            )

    Sum = complex(Sum)
    Yslm = (-1) ** emm * (
        np.sqrt(
            fact(ell + emm)
            * fact(ell - emm)
            * (2 * ell + 1)
            / (
                4
                * np.pi
                * fact(ell + abs_spin_weight)
                * fact(ell - abs_spin_weight)
            )
        )
        * np.sin(theta / 2) ** (2 * ell)
        * Sum
    )

    return factor * Yslm


def Yslm_vec(spin_weight, ell, emm, theta_grid, phi_grid):
    """Spin-weighted spherical harmonics fast evaluations on numpy arrays for vectorized evaluations.

    Inputs
    -----------

    spin_weight :	int
                                    The Spin weight.
    ell :	int
                    The mode number :math:`\\ell'.
    emm :	int
                    The azimuthal mode number :math:`m'.
    theta : float
                    The polar angle  :math:`\\theta` in radians,
    phi :	float
                    The aximuthal angle :math:`\\phi' in radians.

    Returns
    --------
    Yslm :	float
                    The value of Yslm at :math:`\\theta, phi'.

            Note
            ----
            This is accurate upto 14 decimals for L upto 25.
    """

    from math import comb

    fact = np.math.factorial

    theta_grid = np.array(theta_grid)
    phi_grid = np.array(phi_grid)

    Sum = 0 + 1j * 0

    factor = 1
    if spin_weight < 0:
        factor = (-1) ** ell
        theta_grid = np.pi - theta_grid
        phi_grid += np.pi

    abs_spin_weight = abs(spin_weight)

    for aar in range(0, ell - abs_spin_weight + 1):
        subterm = 0

        if (aar + abs_spin_weight - emm) < 0 or (
            ell - aar - abs_spin_weight
        ) < 0:
            message(f"Skipping r {aar}", message_verbosity=3)
            continue
        else:
            term1 = comb(ell - abs_spin_weight, aar)
            term2 = comb(ell + abs_spin_weight, aar + abs_spin_weight - emm)
            term3 = np.power(float(-1), (ell - aar - abs_spin_weight))
            term4 = np.exp(1j * emm * phi_grid)
            term5 = np.power(
                np.tan(theta_grid / 2), (-2 * aar - abs_spin_weight + emm)
            )
            subterm = term1 * term2 * term3 * term4 * term5

            Sum += subterm

    Yslmv = float(-1) ** emm * (
        np.sqrt(
            fact(ell + emm)
            * fact(ell - emm)
            * (2 * ell + 1)
            / (
                4
                * np.pi
                * fact(ell + abs_spin_weight)
                * fact(ell - abs_spin_weight)
            )
        )
        * np.sin(theta_grid / 2) ** (2 * ell)
        * Sum
    )

    return factor * Yslmv


def Yslm_prec(spin_weight, ell, emm, theta, phi, prec=24):
    """Spin-weighted spherical harmonics function with precise computations.
                            Uses a symbolic method evaluated at the degree of precision requested
                            by the user.
    Parameters
    ----------

    spin_weight :	int
                                    The Spin weight.
    ell :	int
                    The mode number :math:`\\ell'.
    emm :	int
                    The azimuthal mode number :math:`m'.
    theta : float
                    The polar angle  :math:`\\theta` in radians,
    phi :	float
                    The aximuthal angle :math:`\\phi' in radians.
    pres : int, optional
               The precision i.e. number of digits to compute
               upto. Default value is 16.
    Returns
    --------
    Yslm :	float
                    The value of Yslm at :math:`\\theta, phi'.

    """
    import sympy as sp

    # tv, pv = theta, phi
    th, ph = sp.symbols("theta phi")

    Yslm_expr = Yslm_prec_sym(spin_weight, ell, emm)

    if spin_weight < 0:
        theta = np.pi - theta
        phi = np.pi + phi

    return Yslm_expr.evalf(
        prec, subs={th: sp.Float(f"{theta}"), ph: sp.Float(f"{phi}")}
    )


def Yslm_prec_sym(spin_weight, ell, emm):
    """Spin-weighted spherical harmonics precise, symbolic computation for deferred evaluations.
       Is dependent on variables th: theta and ph:phi.
    Parameters
    ----------

    spin_weight :	int
                                    The Spin weight.
    ell :	int
                    The mode number :math:`\\ell'.
    emm :	int
                    The azimuthal mode number :math:`m'.
    theta : float
                    The polar angle  :math:`\\theta` in radians,
    phi :	float
                    The aximuthal angle :math:`\\phi' in radians.
    pres : int, optional
               The precision i.e. number of digits to compute
               upto. Default value is 16.

    Returns
    --------
    Yslm :	sym
                    The value of Yslm at :math:`\\theta, phi'.

    """
    import sympy as sp

    th, ph = sp.symbols("theta phi")

    fact = sp.factorial
    Sum = 0

    abs_spin_weight = abs(spin_weight)
    # To get negative spin weight SWSH
    # in terms of positive spin weight
    factor = 1
    if spin_weight < 0:
        factor = sp.Pow(-1, ell)

    for aar in range(ell - abs_spin_weight + 1):
        if (aar + abs_spin_weight - emm) < 0 or (
            ell - aar - abs_spin_weight
        ) < 0:
            # message('Continuing')
            continue
        else:
            # message('r, l, s, m', r, l, s, m)
            # a1 = sp.binomial(ell - spin_weight, aar)
            # message(a1)
            # a2 = sp.binomial(ell + spin_weight, aar + spin_weight - emm)
            # message(a2)
            # a3 = sp.exp(1j * emm * phi)
            # message(a3)
            # a4 = sp.tan(theta / 2)
            # message(a4)

            Sum += (
                sp.binomial(ell - abs_spin_weight, aar)
                * sp.binomial(
                    ell + abs_spin_weight, aar + abs_spin_weight - emm
                )
                * sp.Pow((-1), (ell - aar - abs_spin_weight))
                * sp.exp(sp.I * emm * ph)
                * sp.Pow(sp.cot(th / 2), (2 * aar + abs_spin_weight - emm))
            )

    Yslm_expr = sp.Pow(-1, emm) * (
        sp.sqrt(
            fact(ell + emm)
            * fact(ell - emm)
            * (2 * ell + 1)
            / (
                4
                * sp.pi
                * fact(ell + abs_spin_weight)
                * fact(ell - abs_spin_weight)
            )
        )
        * sp.Pow(sp.sin(th / 2), (2 * ell))
        * Sum
    )

    Yslm_expr = factor * sp.simplify(Yslm_expr)

    return Yslm_expr


def rotate_polarizations(wf, alpha):
    """Rotate the polarizations of the time domain
    observer waveform by :math:`2\alpha`

    Parameters
    ----------
    wf : 1d array
             The complex observer waveform to rotate.
    alpha : float
                    The coordinate angle to rotate the polarizations
                    in radians. Note that the polarizarions would
                    rotate by :math:`2 \alpha` on a cordinate
                    rotation of :math:`\alpha`.

    Returns
    -------
    rot_wf : 1d array
                     The rotated waveform.
    """

    h1, h2 = wf.real, wf.imag

    rh1 = np.cos(2 * alpha) * h1 - np.sin(2 * alpha) * h2
    rh2 = np.sin(2 * alpha) * h1 + np.cos(2 * alpha) * h2

    return rh1 + 1j * rh2
